#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#
from .TranslatorOutputType import TranslatorOutputType
from .internal._GXCommon import _GXCommon
# pylint: disable=too-many-arguments, too-many-instance-attributes
class GXDLMSTranslatorStructure:
    """
    This class is used internally in GXDLMSTranslator to save generated XML.
    """

    def setOffset(self, value):
        """
        Amount of spaces.
        """
        if value < 0:
            raise ValueError("offset")
        self.offset = value

    #
    # Constructor.
    #
    # @param list
    #            List of tags.
    #
    def __init__(self, type_, numericsAshex, numericshex, isHex, addComments, list_):
        self.sb = ""
        self.tags = None
        self.outputType = None
        # Are numeric values shows as hex.
        self.showNumericsAsHex = False
        # Amount of spaces.
        self.offset = 0
        self.outputType = type_
        numericsAshex = numericshex
        self.showNumericsAsHex = numericsAshex
        self.showStringAsHex = isHex
        self.tags = list_
        # Are comments added.
        self.comments = addComments
        self.omitNameSpace = False

    def __str__(self):
        return str(self.sb)

    def getDataType(self, type_):
        return self.__getTag(_GXCommon.DATA_TYPE_OFFSET + type_)

    def __getTag(self, tag):
        if self.outputType == TranslatorOutputType.SIMPLE_XML or self.omitNameSpace:
            return self.tags.get(tag)
        return "x:" + self.tags.get(tag)

    #
    # Append spaces to the buffer.
    #
    # @param count
    #            Amount of spaces.
    #
    def appendSpaces(self, count):
        pos = 0
        while pos != count:
            self.sb += ' '
            pos += 1

    def appendLine(self, tag, name=None, value=None):
        self.appendSpaces(2 * self.offset)
        if name is None and value is None:
            self.sb += tag
        else:
            self.sb += '<'
            if isinstance(tag, int):
                tag = self.__getTag(tag)
            self.sb += tag
            if self.outputType == TranslatorOutputType.SIMPLE_XML:
                self.sb += ' '
                if name is None:
                    self.sb += "Value"
                else:
                    self.sb += name
                self.sb += "=\""
            else:
                self.sb += '>'
            self.sb += value
            if self.outputType == TranslatorOutputType.SIMPLE_XML:
                self.sb += "\" />"
            else:
                self.sb += "</"
                self.sb += tag
                self.sb += '>'
        self.sb += '\r'
        self.sb += '\n'

    #
    # Append comment.
    #
    # @param comment
    #            Comment to add.
    #
    def appendComment(self, comment):
        if self.comments:
            self.appendSpaces(2 * self.offset)
            self.sb += "<!--"
            self.sb += comment
            self.sb += "-->"
            self.sb += '\r'
            self.sb += '\n'

    #
    # Start comment section.
    #
    # @param comment
    #            Comment to add.
    #
    def startComment(self, comment):
        if self.comments:
            self.appendSpaces(2 * self.offset)
            self.sb += "<!--"
            self.sb += comment
            self.sb += '\r'
            self.sb += '\n'
            self.offset += 1

    #
    # End comment section.
    #
    def endComment(self):
        if self.comments:
            self.offset -= 1
            self.appendSpaces(2 * self.offset)
            self.sb += "-->"
            self.sb += '\r'
            self.sb += '\n'

    def append(self, tag, start=None):
        if isinstance(tag, str):
            self.sb += tag
            return
        if start:
            self.appendSpaces(2 * self.offset)
            self.sb += '<'
        else:
            self.sb += "</"
        self.sb += self.__getTag(tag)
        self.sb += '>'

    def appendStartTag(self, tag, name=None, value=None, plain=False):
        if isinstance(name, int):
            tag = self.__getTag(tag << 8 | name)
            name = None
        elif isinstance(tag, int):
            tag = self.__getTag(tag)
        self.appendSpaces(2 * self.offset)
        self.sb += '<'
        self.sb += tag
        if self.outputType == TranslatorOutputType.SIMPLE_XML and name:
            self.sb += ' '
            self.sb += name
            self.sb += "=\""
            self.sb += value
            self.sb += "\" >"
        else:
            self.sb += '>'
        if not plain:
            self.sb += '\r'
            self.sb += '\n'
        self.offset += 1

    def appendEndTag(self, tag, plain=False):
        if isinstance(tag, int):
            if not isinstance(plain, bool) and (isinstance(plain, int)):
                tag = self.__getTag(tag << 8 | plain)
            else:
                tag = self.__getTag(tag)
        self.setOffset(self.offset - 1)
        if not isinstance(plain, bool) or not plain:
            self.appendSpaces(2 * self.offset)
        self.sb += "</"
        self.sb += tag
        self.sb += '>'
        self.sb += '\r'
        self.sb += '\n'

    def appendEmptyTag(self, tag):
        if isinstance(tag, int):
            tag = self.__getTag(tag)
        self.appendSpaces(2 * self.offset)
        self.sb += "<"
        self.sb += tag
        self.sb += '/'
        self.sb += '>'
        self.sb += '\r'
        self.sb += '\n'

    def trim(self):
        """
        Remove \r\n.
        """
        self.sb = self.sb[:-2]

    def getXmlLength(self):
        """
        XML Length.
        """
        return len(self.sb)

    def setXmlLength(self, value):
        """
        Set XML Length.
        """
        self.sb = self.sb[0:value]

    #
    # Convert integer to string.
    #
    # @param value
    #            Converted value.
    # @param desimals
    #            Amount of decimals.
    # @param forceHex
    #            Force value as hex.
    # Integer as string.
    #
    def integerToHex(self, value, desimals, forceHex=False):
        if forceHex or (self.showNumericsAsHex and self.outputType == TranslatorOutputType.SIMPLE_XML):
            return _GXCommon.integerToHex(value, desimals)
        return str(value)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import datetime
import calendar
import time
from .enums import DateTimeSkips, ClockStatus, DateTimeExtraInfo
from .GXTimeZone import GXTimeZone

###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class
class GXDateTime:
    def __init__(self, value=None, pattern=None):
        """
        Constructor.

        value: Date-time value.
        pattern: Date-time pattern that is used when value is a string.
        """
        self.extra = DateTimeExtraInfo.NONE
        self.skip = DateTimeSkips.NONE
        self.status = ClockStatus.OK
        self.dayOfWeek = 0
        if isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                if time.localtime().tm_isdst:
                    #If DST is used.
                    tz = GXTimeZone(int(-time.altzone / 60))
                else:
                    tz = GXTimeZone(int(-time.timezone / 60))
                self.value = datetime.datetime(value.year, value.month, value.day, value.hour, value.minute, value.second, 0, tzinfo=tz)
            else:
                self.value = value
        elif isinstance(value, str):
            self.value = self.fromString(value, pattern)
        elif isinstance(value, GXDateTime):
            self.value = value.value
            self.skip = value.skip
            self.extra = value.extra
        elif not value:
            self.value = None
        else:
            raise ValueError("Invalid datetime value.")

    @classmethod
    def __isNumeric(cls, value):
        return '0' <= value <= '9'

    @classmethod
    def __get_pattern(cls, loading):
        tm = datetime.datetime(1901, 2, 3, 13, 14, 15)
        pm = tm.strftime('%p')
        date = tm.strftime('%x')
        for s in date:
            if s < '0' or s > '9':
                sep = s
                break
        dp = ""
        tp = ""
        appendPM = ""
        date = date.replace(" ", sep).replace(" ", "")
        for d in date.split(sep):
            if not cls.__isNumeric(d):
                if d == '' and sep != '.':
                    d = '.'
                appendPM += d
                continue
            dp += sep
            tmp = int(d)
            if tmp in (1901, 1):
                if loading:
                    dp += "%Y"
                elif len(d) == 2:
                    dp += "%y"
                elif len(d) == 1:
                    dp += "%-y"
                else:
                    dp += "%Y"
            elif tmp == 2:
                if len(d) == 2 or loading:
                    dp += "%m"
                else:
                    dp += "%-m"
            elif tmp == 3:
                if len(d) == 2 or loading:
                    dp += "%d"
                else:
                    dp += "%-d"

        date = tm.strftime('%X').replace(pm, "").replace(" ", "")
        for s in date:
            if s < '0' or s > '9':
                sep = s
                break
        for t in date.split(sep):
            tp += sep
            tmp = int(t)
            if tmp in (13, 1):
                if len(t) == 2 or loading:
                    tp += "%H"
                else:
                    tp += "%-H"
            elif tmp == 14:
                if len(t) == 2 or loading:
                    tp += "%M"
                else:
                    tp += "%-M"
            elif tmp == 15:
                if len(t) == 2 or loading:
                    tp += "%S"
                else:
                    tp += "%-S"
        if not pm or tm.strftime('%c').find(pm) == -1:
            return dp[1:] + appendPM + " " + tp[1:]
        tp = tp.replace("H", "I")
        if not appendPM:
            return dp[1:] + " " + tp[1:] + " " + pm
        return dp[1:] + " " + appendPM + " " + tp[1:]

    #Check is time zone included and return index of time zone.
    @classmethod
    def __timeZonePosition(cls, value):
        if len(value) > 5:
            pos = len(value) - 6
            sep = value[pos]
            if sep in('-', '+'):
                return pos
            if value[len(value) - 1] == 'Z':
                return len(value) - 1
        return -1

    #
    # Constructor
    #
    # @param value
    #            Date time value as a string.
    #pylint: disable=too-many-nested-blocks
    def fromString(self, value, pattern=None):
        if self.skip is None:
            self.skip = DateTimeSkips.NONE
        if self.status is None:
            self.status = ClockStatus.OK
        if self.extra is None:
            self.extra = DateTimeExtraInfo.NONE
        if value:
            if not pattern:
                pattern = self.__get_pattern(True)
            pattern = self._remove(pattern)
            if value.find('BEGIN') != -1:
                self.extra |= DateTimeExtraInfo.DST_BEGIN
                value = value.replace("BEGIN", "01")
            if value.find("END") != -1:
                self.extra |= DateTimeExtraInfo.DST_END
                value = value.replace("END", "01")
            if value.find("LASTDAY2") != -1:
                self.extra |= DateTimeExtraInfo.LAST_DAY2
                value = value.replace("LASTDAY2", "01")
            if value.find("LASTDAY") != -1:
                self.extra |= DateTimeExtraInfo.LAST_DAY
                value = value.replace("LASTDAY", "01")
            v = value
            #Time zone is not added to time or date objects.
            addTimeZone = type(self).__name__ == GXDateTime.__name__
            if value.find('*') != -1:
                lastFormatIndex = -1
                pos = 0
                while pos < len(value):
                    c = value[pos]
                    if not self.__isNumeric(c):
                        if c == '*':
                            end = lastFormatIndex + 1
                            c = pattern[end]
                            while end + 1 < len(pattern) and pattern[end] == c:
                                end += 1
                            if pattern[end] == 'Y':
                                v = str(v[0:pos]) + "Y" + str(v[pos + 1:])
                            elif pattern[end] == '%-Y':
                                v = str(v[0:pos]) + "Y" + str(v[pos + 1:])
                            elif pattern[end] == '%Y':
                                v = str(v[0:pos]) + "Y" + str(v[pos + 1:])
                            else:
                                v = str(v[0:pos]) + "1" + str(v[pos + 1:])
                            tmp = pattern[lastFormatIndex + 1: end + 1].strip()
                            if tmp in ("%y", "%-y", "%Y", "%-Y"):
                                addTimeZone = False
                                self.skip |= DateTimeSkips.YEAR
                            elif tmp in ("%m", "%-m"):
                                addTimeZone = False
                                self.skip |= DateTimeSkips.MONTH
                            elif tmp in ("%d", "%-d"):
                                addTimeZone = False
                                self.skip |= DateTimeSkips.DAY
                            elif tmp in ("%H", "%-H"):
                                addTimeZone = False
                                self.skip |= DateTimeSkips.HOUR
                                pos2 = pattern.find("%p")
                                if pos2 != -1:
                                    pattern.replace(pos2, pos2 + 1, "")
                            elif tmp in ("%M", "%-M"):
                                addTimeZone = False
                                self.skip |= DateTimeSkips.MINUTE
                            elif tmp in ("%S", "%-S"):
                                self.skip |= DateTimeSkips.SECOND
                            elif tmp and not tmp == "G":
                                raise ValueError("Invalid date time format.")
                        else:
                            lastFormatIndex = pattern.find(str(c), lastFormatIndex + 1)
                    pos += 1
                v = v.replace("Y", "2000")
            self.skip |= DateTimeSkips.DAY_OF_WEEK | DateTimeSkips.MILLISECOND
            #If time zone is used.
            if addTimeZone:
                pos = self.__timeZonePosition(v)
                tz = None
                if pos != -1:
                    if v[pos] != 'Z':
                        tz = 60 * int(v[pos + 1:pos + 3]) + int(v[pos + 4:])
                    else:
                        tz = 0
                    v = v[0:pos]
                tmp = datetime.datetime.strptime(v, pattern)
                if tz is None:
                    tmp = datetime.datetime(tmp.year, tmp.month, tmp.day, tmp.hour, tmp.minute, tmp.second, 0)
                else:
                    tmp = datetime.datetime(tmp.year, tmp.month, tmp.day, tmp.hour, tmp.minute, tmp.second, 0, tzinfo=GXTimeZone(tz))
                return tmp
            return datetime.datetime.strptime(v.strip(), pattern)
        return None

     #pylint: disable=no-self-use
    def _remove(self, format_):
        #Do nothing.
        return format_

    def toFormatString(self, pattern=None):
        return self.__toFormatString(True, pattern)

    def toFormatMeterString(self, pattern=None):
        return self.__toFormatString(False, pattern)

    @classmethod
    def __toLocal(cls, value):
        #Convert current time to local.
        if value.tzinfo is None:
            #If meter is not use time zone.
            return value
        timestamp = calendar.timegm(value.utctimetuple())
        local_dt = datetime.datetime.fromtimestamp(timestamp)
        assert value.resolution >= datetime.timedelta(microseconds=1)
        return local_dt.replace(microsecond=value.microsecond)

    def __getTimeZone(self):
        if self.value.tzinfo:
            return self.value.tzname()
        return ""
#        if self.value.utcoffset() is None:
#            str_ = ""
#        else:
#
#            offset = int(self.value.utcoffset().seconds / 60)
#            if offset == 0:
#                str_ = "Z"
#            else:
#                if offset > 0:
#                    str_ = "+"
#                else:
#                    str_ = "-"
#                str_ += str(int(offset / 60)).zfill(2)
#                str_ += ":"
#                str_ += str(offset % 60).zfill(2)#
#        return str_

    def __toFormatString(self, useLocalTime, pattern=None):
        if not self.value:
            return ""
        if self.skip != DateTimeSkips.NONE:
            #  Separate date and time parts.
            if not pattern:
                pattern = self.__get_pattern(True)
            pattern = self._remove(pattern)

            if self.extra & DateTimeExtraInfo.DST_BEGIN != 0:
                pattern = self._replace(pattern, "%m", "BEGIN")
                pattern = self._replace(pattern, "%-m", "BEGIN")
            elif self.extra & DateTimeExtraInfo.DST_END != 0:
                pattern = self._replace(pattern, "%m", "END")
                pattern = self._replace(pattern, "%-m", "END")
            elif self.extra & DateTimeExtraInfo.LAST_DAY != 0:
                pattern = self._replace(pattern, "%d", "LASTDAY")
                pattern = self._replace(pattern, "%-d", "LASTDAY")
            elif self.extra & DateTimeExtraInfo.LAST_DAY2 != 0:
                pattern = self._replace(pattern, "%d", "LASTDAY2")
                pattern = self._replace(pattern, "%-d", "LASTDAY2")

            if self.skip & DateTimeSkips.YEAR != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%y", "*")
                pattern = self._replace(pattern, "%-y", "*")
                pattern = self._replace(pattern, "%-Y", "*")
                pattern = self._replace(pattern, "%Y", "*")
            if self.skip & DateTimeSkips.MONTH != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%m", "*")
                pattern = self._replace(pattern, "%-m", "*")
            if self.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%d", "*")
                pattern = self._replace(pattern, "%-d", "*")
            if self.skip & DateTimeSkips.HOUR != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%H", "*")
                pattern = self._replace(pattern, "%-H", "*")
                pattern = self._replace(pattern, "%I", "*")
                pattern = self._replace(pattern, "%-I", "*")
                pattern = self._remove_(pattern, "p", False)
            if self.skip & DateTimeSkips.MILLISECOND != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%f", "*")
            else:
                index = pattern.find("%S")
                if index != -1:
                    sep = pattern[index - 1]
                    pattern.replace("%S", "%S" + sep + "%f")
                else:
                    index = pattern.find("%-S")
                    if index != -1:
                        sep = pattern[index - 1]
                        pattern.replace("%-S", "%-S" + sep + "%f")
            if self.skip & DateTimeSkips.SECOND != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%S", "*")
                pattern = self._replace(pattern, "%-S", "*")
            else:
                index = pattern.find("%M")
                if index != -1:
                    sep = pattern[index - 1]
                    pattern.replace("%M", "%M" + sep + "%S")
            if self.skip & DateTimeSkips.MINUTE != DateTimeSkips.NONE:
                pattern = self._replace(pattern, "%M", "*")
                pattern = self._replace(pattern, "%-M", "*")

            if useLocalTime:
                return self.__toLocal(self.value).strftime(pattern)
            return self.value.strftime(pattern) + self.__getTimeZone()
        if pattern:
            if useLocalTime:
                return self.__toLocal(self.value).strftime(pattern)
            return self.value.strftime(pattern) + self.__getTimeZone()
        if useLocalTime:
            return self.__toLocal(self.value).strftime("%x %X")
        return self.value.strftime(self.__get_pattern(True)) + self.__getTimeZone()

    @classmethod
    def _remove_(cls, value, tag, removeSeparator):
        pos = value.find(tag)
        if pos != -1:
            len_ = pos + len(tag)
            if pos != 0 and removeSeparator:
                pos -= 1
            value = value[0:pos] + value[len_:]
        return value

    @classmethod
    def _replace(cls, value, tag, replacement):
        pos = value.find(tag)
        if pos != -1:
            value = value.replace(tag, replacement)
        return value

    def __str__(self):
        #Returns date-time of the meter using local time zone.
        return self.__toString(True)

    def toString(self, pattern=None):
        return self.__toString(True, pattern)

    def toMeterString(self, pattern=None):
        #Returns date-time of the meter using meter time zone.
        return self.__toString(False, pattern)

    def __toString(self, useLocalTime, pattern=None):
        if not self.value:
            return ""
        if self.skip != DateTimeSkips.NONE:
            #  Separate date and time parts.
            if not pattern:
                pattern = self.__get_pattern(False)
            pattern = self._remove(pattern)
            if self.skip & DateTimeSkips.YEAR != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%Y", True)
                pattern = self._remove_(pattern, "%y", True)
                pattern = self._remove_(pattern, "%-y", True)
                pattern = self._remove_(pattern, "%-Y", True)
            if self.skip & DateTimeSkips.MONTH != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%m", True)
                pattern = self._remove_(pattern, "%-m", True)
            if self.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%d", True)
                pattern = self._remove_(pattern, "%-d", True)
            if self.skip & DateTimeSkips.HOUR != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%H", True)
                pattern = self._remove_(pattern, "%-H", True)
                pattern = self._remove_(pattern, "%p", True)
            if self.skip & DateTimeSkips.MILLISECOND != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%f", True)
            else:
                index = pattern.find("%S")
                if index != -1:
                    sep = pattern[index - 1]
                    pattern = pattern.replace("%S", "%S" + sep + "%f")
                else:
                    index = pattern.find("%-S")
                    if index != -1:
                        sep = pattern[index - 1]
                        pattern = pattern.replace("%-S", "%-S" + sep + "%f")
            if self.skip & DateTimeSkips.SECOND != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%S", True)
                pattern = self._remove_(pattern, "%-S", True)
            else:
                index = pattern.find("%M")
                if index == -1:
                    index = pattern.find("%-M")
                    if index != -1:
                        sep = pattern[index - 1]
                        pattern.replace("%-M", "%-M" + sep + "-S")
                    else:
                        sep = pattern[index - 1]
                        pattern.replace("%M", "%M" + sep + "S")
            if self.skip & DateTimeSkips.MINUTE != DateTimeSkips.NONE:
                pattern = self._remove_(pattern, "%M", True)
                pattern = self._remove_(pattern, "%-M", True)
            if useLocalTime:
                return self.__toLocal(self.value).strftime(pattern.strip())
            return self.value.strftime(pattern.strip()) + self.__getTimeZone()
        if pattern:
            if useLocalTime:
                return self.__toLocal(self.value).strftime(pattern)
            return self.value.strftime(pattern) + self.__getTimeZone()
        if useLocalTime:
            return self.__toLocal(self.value).strftime("%x %X")
        return self.value.strftime("%x %X") + self.__getTimeZone()

    #
    # Get difference between given time and run time in ms.
    #
    # @param start
    #            Start date time.
    # @param to
    #            Compared time.
    # Difference in milliseconds.
    #
    @classmethod
    def getDifference(cls, start, to):
        diff = 0
        cal = to.getLocalCalendar()
        #  Compare seconds.
        if not to.skip & DateTimeSkips.SECOND != DateTimeSkips.NONE:
            if start.second < cal.second:
                diff += (cal.second - start.second) * 1000
            else:
                diff -= (start.second - cal.second) * 1000
        elif diff < 0:
            diff = 60000 + diff
        #  Compare minutes.
        if not to.skip & DateTimeSkips.MINUTE != DateTimeSkips.NONE:
            if start.minute < cal.minute:
                diff += (cal.minute - start.minute) * 60000
            else:
                diff -= (start.minute - cal.minute) * 60000
        elif diff < 0:
            diff = 60 * 60000 + diff
        #  Compare hours.
        if not to.skip & DateTimeSkips.HOUR != DateTimeSkips.NONE:
            if start.hour < cal.hour:
                diff += (cal.hour - start.hour) * 60 * 60000
            else:
                diff -= (start.hour - cal.hour) * 60 * 60000
        elif diff < 0:
            diff = 60 * 60000 + diff
        #  Compare days.
        if not to.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
            if start.day < cal.day:
                diff += (cal.month - start.month) * 24 * 60 * 60000
            elif start.month != cal.month:
                if not to.skip & DateTimeSkips.DAY != DateTimeSkips.NONE:
                    diff += (cal.month - start.month) * 24 * 60 * 60000
                else:
                    diff = ((cls.daysInMonth(start.year, start.month) - start.day + cal.day) * 24 * 60 * 60000) + diff
        elif diff < 0:
            diff = 24 * 60 * 60000 + diff
        #  Compare months.
        if not to.skip & DateTimeSkips.MONTH != DateTimeSkips.NONE:
            if start.month < cal.month:
                m = start.day
                while m != cal.day:
                    diff += cls.daysInMonth(start.year, m) * 24 * 60 * 60000
                    m += 1
            else:
                m = cal.day
                while m != start.day:
                    diff -= cls.daysInMonth(start.year, m) * 24 * 60 * 60000
                    m += 1
        elif diff < 0:
            diff = cls.daysInMonth(start.year, start.month) * 24 * 60 * 60000 + diff
        return diff

    #
    # Get the number of days in that month.
    #
    # year
    # Year.
    # month
    # Month.
    # Number of days in month.
    #
    @classmethod
    def daysInMonth(cls, year, month):
        return calendar.monthrange(year, month)

    #
    # Get date time from Epoch time.
    #
    # @param unixTime
    #            Unix time.
    # Date and time.
    #
    @classmethod
    def fromUnixTime(cls, unixTime):
        return GXDateTime(datetime.datetime(unixTime * 1000))

    #
    # Convert date time to Epoch time.
    #
    # @param value
    #            Date and time.
    # Returns unix time.
    #
    @classmethod
    def toUnixTime(cls, value):
        if isinstance(value, (datetime.datetime)):
            return value.utctimetuple()
        if isinstance(value, (GXDateTime)):
            return value.value.utctimetuple()
        return int(value.value / 1000)

    #
    # Get date time from high resolution clock time.
    #
    # highResolution:  High resolution clock time is milliseconds since 1970-01-01 00:00:00.
    # Returns Date and time.
    #
    @classmethod
    def fromHighResolutionTime(cls, unixTime):
        return GXDateTime(datetime.datetime(unixTime))

    #
    # Convert date time to high resolution time.
    #
    # @param value
    #            Date and time.
    # Returns high resolution time.
    #
    @classmethod
    def toHighResolutionTime(cls, value):
        if isinstance(value, datetime.datetime):
            return value.utctimetuple() * 1000.0
        if isinstance(value, GXDateTime):
            return value.value.utctimetuple() * 1000.0
        return int(value.value)

    #
    # Get date time as hex string.
    #
    # @param addSpace
    #           Add space between bytes.
    # @param useMeterTimeZone
    #           Date-Time values are shown using meter's time zone and it's not localized to use PC time.</param>
    # Returns date time as a hex string.
    #
    def toHex(self, addSpace, useMeterTimeZone):
        #pylint: disable=import-outside-toplevel
        from .GXByteBuffer import GXByteBuffer
        from .enums import DataType
        from .internal._GXCommon import _GXCommon
        from .GXDLMSSettings import GXDLMSSettings
        buff = GXByteBuffer()
        settings = GXDLMSSettings(False)
        settings.UseUtc2NormalTime = useMeterTimeZone
        _GXCommon.setData(settings, buff, DataType.OCTET_STRING, self)
        #Dont add data type or length.
        return buff.toHex(addSpace, 2)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSScript import GXDLMSScript
from .GXDLMSScriptAction import GXDLMSScriptAction
from .enums import ScriptActionType

# pylint: disable=too-many-instance-attributes
class GXDLMSScriptTable(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSScriptTable
    """
    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        super(GXDLMSScriptTable, self).__init__(ObjectType.SCRIPT_TABLE, ln, sn)
        self.scripts = list()

    def getValues(self):
        return [self.logicalName,
                self.scripts]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Scripts
        if all_ or not self.isRead(2):
            attributes.append(2)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 2

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            cnt = len(self.scripts)
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            #  Add count
            _GXCommon.setObjectCount(cnt, data)
            for it in self.scripts:
                data.setUInt8(DataType.STRUCTURE)
                #  Count
                data.setUInt8(2)
                #  Script_identifier:
                _GXCommon.setData(settings, data, DataType.UINT16, it.id)
                data.setUInt8(DataType.ARRAY)
                #  Count
                data.setUInt8(len(it.actions))
                for a in it.actions:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(5)
                    #  service_id
                    _GXCommon.setData(settings, data, DataType.ENUM, a.type_)
                    if not a.target:
                        #  class_id
                        _GXCommon.setData(settings, data, DataType.UINT16, a.objectType)
                        #  logical_name
                        _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(a.logicalName))
                    else:
                        #  class_id
                        _GXCommon.setData(settings, data, DataType.UINT16, a.target.objectType)
                        _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(a.target.logicalName))
                    _GXCommon.setData(settings, data, DataType.INT8, a.index)
                    _GXCommon.setData(settings, data, a.parameterDataType, a.parameter)
            return data
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def setValue(self, settings, e):
        # pylint: disable=import-outside-toplevel,bad-option-value,too-many-nested-blocks,redefined-variable-type
        from .._GXObjectFactory import _GXObjectFactory
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.scripts = []
            if isinstance(e.value, list) and e.value:
                if isinstance(e.value[0], list):
                    for item in e.value:
                        script = GXDLMSScript()
                        script.id = item[0]
                        self.scripts.append(script)
                        for arr in item[1]:
                            it = GXDLMSScriptAction()
                            it.type_ = arr[0]
                            ot = arr[1]
                            ln = _GXCommon.toLogicalName(arr[2])
                            t = settings.objects.findByLN(ot, ln)
                            if t is None:
                                t = _GXObjectFactory.createObject(ot)
                                t.logicalName = ln
                            it.target = t
                            it.index = arr[3]
                            it.parameter = arr[4]
                            it.parameterDataType = _GXCommon.getDLMSDataType(it.parameter)
                            script.actions.append(it)
                else:
                    script = GXDLMSScript()
                    script.id = e.value[0]
                    arr = e.value[1]
                    it = GXDLMSScriptAction()
                    it.type_ = arr[0]
                    ot = arr[1]
                    ln = _GXCommon.toLogicalName(arr[2])
                    t = settings.objects.findByLN(ot, ln)
                    if t is None:
                        t = _GXObjectFactory.createObject(ot)
                        t.logicalName = ln
                    it.target = t
                    it.index = arr[3]
                    it.parameter = arr[4]
                    script.actions.append(it)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def execute(self, client, script):
        if isinstance(script, GXDLMSScript):
            script = script.id
        return client.method(self, 1, script, DataType.UINT16)

    def load(self, reader):
        #pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory
        self.scripts = []
        if reader.isStartElement("Scripts", True):
            while reader.isStartElement("Script", True):
                it = GXDLMSScript()
                self.scripts.append(it)
                it.id = reader.readElementContentAsInt("ID")
                if reader.isStartElement("Actions", True):
                    while reader.isStartElement("Action", True):
                        a = GXDLMSScriptAction()
                        a.type_ = ScriptActionType(reader.readElementContentAsInt("Type"))
                        ot = ObjectType(reader.readElementContentAsInt("ObjectType"))
                        ln = reader.readElementContentAsString("LN")
                        t = reader.objects.findByLN(ot, ln)
                        if t is None:
                            t = _GXObjectFactory.createObject(ot)
                            t.logicalName = ln
                        a.target = t
                        a.index = reader.readElementContentAsInt("Index")
                        a.parameterDataType = reader.readElementContentAsInt("ParameterDataType")
                        a.parameter = reader.readElementContentAsObject("Parameter", None)
                        it.actions.append(a)
                    reader.readEndElement("Actions")
            reader.readEndElement("Scripts")

    def save(self, writer):
        if self.scripts:
            writer.writeStartElement("Scripts")
            for it in self.scripts:
                writer.writeStartElement("Script")
                writer.writeElementString("ID", it.id)
                writer.writeStartElement("Actions")
                for a in it.actions:
                    writer.writeStartElement("Action")
                    writer.writeElementString("Type", int(a.type_))
                    if a.target is None:
                        writer.writeElementString("ObjectType", int(ObjectType.NONE))
                        writer.writeElementString("LN", "0.0.0.0.0.0")
                        writer.writeElementString("Index", "0")
                        writer.writeElementString("ParameterDataType", int(DataType.NONE))
                        writer.writeElementObject("Parameter", "")
                    else:
                        writer.writeElementString("ObjectType", int(a.target.objectType))
                        writer.writeElementString("LN", a.target.logicalName)
                        writer.writeElementString("Index", a.index)
                        writer.writeElementString("ParameterDataType", int(a.parameterDataType))
                        writer.writeElementObject("Parameter", a.parameter)
                    writer.writeEndElement()
                writer.writeEndElement()
                writer.writeEndElement()
            writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType

# pylint: disable=too-many-instance-attributes
class GXDLMSData(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSData
    """

    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.DATA, ln, sn)
        self.value = None

    def getValues(self):
        return [self.logicalName,
                self.value]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Value
        if all_ or self.canRead(2):
            attributes.append(2)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 2

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            #pylint: disable=super-with-arguments
            return super(GXDLMSData, self).getDataType(index)
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.value
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.value = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.value = reader.readElementContentAsObject("Value", None, self, 2)

    def save(self, writer):
        writer.writeElementObject("Value", self.value, self.getDataType(2), self.getUIDataType(2))



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .Access import Access
from .AccessMode import AccessMode
from .AccessServiceCommandType import AccessServiceCommandType
from .ApplicationReference import ApplicationReference
from .AssociationResult import AssociationResult
from .Authentication import Authentication
from .BerType import BerType
from .ClockStatus import ClockStatus
from .Command import Command
from .Conformance import Conformance
from .DataType import DataType
from .DateTimeSkips import DateTimeSkips
from .DateTimeExtraInfo import DateTimeExtraInfo
from .Definition import Definition
from .ErrorCode import ErrorCode
from .ExceptionServiceError import ExceptionServiceError
from .HardwareResource import HardwareResource
from .Initiate import Initiate
from .InterfaceType import InterfaceType
from .KeyAgreement import KeyAgreement
from .LoadDataSet import LoadDataSet
from .MethodAccessMode import MethodAccessMode
from .ObjectType import ObjectType
from .Priority import Priority
from .RequestTypes import RequestTypes
from .Security import Security
from .ServiceClass import ServiceClass
from .ServiceError import ServiceError
from .SourceDiagnostic import SourceDiagnostic
from .Standard import Standard
from .StateError import StateError
from .Task import Task
from .Unit import Unit
from .VdeStateError import VdeStateError
from .HdlcFrameType import HdlcFrameType
from .PduType import PduType
from .Service import Service
from .AcseServiceProvider import AcseServiceProvider



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.GXIntEnum import GXIntEnum

class Ip4SetupIpOptionType(GXIntEnum):
    #pylint: disable=too-few-public-methods

    # If this option is present, the device shall be allowed to send security,
    # compartmentation, handling restrictions and TCC =closed user group)
    # parameters within its IP Datagrams.  The value of the IP-Option- Length
    # Field must be 11, and the IP-Option-Data shall contain the value of the
    # Security, Compartments, Handling Restrictions and Transmission Control
    # code values, as specified in STD0005 / RFC791.
    SECURITY = 0x82

    # If this option is present, the device shall supply routing information
    # to
    # be used by the gateways in forwarding the datagram to the destination,
    # and to record the route information.  The IP-Option-length and
    # IP-Option-Data values are specified in STD0005 / RFC 791.
    LOOSE_SOURCE_AND_RECORD_ROUTE = 0x83

    # If this option is present, the device shall supply routing information
    # to
    # be used by the gateways in forwarding the datagram to the destination,
    # and to record the route information.  The IP-Option-length and
    # IP-Option-Data values are specified in STD0005 / RFC 791.
    STRICT_SOURCE_AND_RECORD_ROUTE = 0x89

    # If this option is present, the device shall as well: send originated IP
    # Datagrams with that option, providing means to record the route of these
    # Datagrams; as a router, send routed IP Datagrams with the route option
    # adjusted according to this option.  The IP-Option-length and
    # IP-Option-Data values are specified in STD0005 / RFC 791.
    #
    RECORD_ROUTE = 0x07

    # If this option is present, the device shall as well: send originated IP
    # Datagrams with that option, providing means to time-stamp the datagram
    # in
    # the route to its destination; as a router, send routed IP Datagrams with
    # the time-stamp option adjusted according to this option.  The
    # IP-Option-length and IP-Option-Data values are specified in STD0005 /
    # RFC 791.
    INTERNET_TIMESTAMP = 0x44



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import xml.etree.cElementTree as ET
from xml.dom import minidom
from .GXDLMSObject import GXDLMSObject
from ..enums import ObjectType
from .GXXmlWriter import GXXmlWriter
from .GXXmlReader import GXXmlReader
from ..GXDLMSConverter import GXDLMSConverter

class GXDLMSObjectCollection(list):
    """
    Collection of DLMS objects.
    """

    #
    # Constructor.
    #
    # forParent: Parent object.
    #
    def __init__(self, forParent=None):
        #pylint: disable=super-with-arguments
        super(GXDLMSObjectCollection, self).__init__()
        self.parent = forParent

    def append(self, item):
        if not isinstance(item, GXDLMSObject):
            raise TypeError('item is not of type GXDLMSObject')
        #pylint: disable=super-with-arguments
        super(GXDLMSObjectCollection, self).append(item)
        item.parent = self

    def extend(self, items):
        #pylint: disable=super-with-arguments
        if not isinstance(items, GXDLMSObjectCollection):
            raise TypeError('items is not of type GXDLMSObjectCollection')
        for it in items:
            super(GXDLMSObjectCollection, self).append(it)
            it.parent = self

    def getObjects(self, type_):
        if isinstance(type_, (int, ObjectType)):
            type_ = [type_]
        items = GXDLMSObjectCollection()
        for it in self:
            if it.objectType in type_:
                items.append(it)
        return items

    def findByLN(self, type_, ln):
        for it in self:
            if type_ in (ObjectType.NONE, it.objectType) and it.logicalName.strip() == ln:
                return it
        return None

    def findBySN(self, sn):
        for it in self:
            if it.shortName == sn:
                return it
        return None

    def __str__(self):
        str_ = '['
        for it in self:
            if str_:
                str_ += ", "
            str_ += it
        str_ += ']'
        return str_

    @classmethod
    def load(cls, file_):
        #pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory
        obj = None
        target = None
        type_ = None
        reader = GXXmlReader(file_, GXDLMSObjectCollection())
        reader.getNext()
        while not reader.isEOF():
            if reader.isStartElement():
                target = reader.name()
                if "Objects".lower() == target.lower():
                    reader.getNext()
                elif target.find("GXDLMS") == 0:
                    try:
                        type_ = GXDLMSConverter.valueOfObjectType(target[6:])
                    except Exception:
                        # pylint:disable=raise-missing-from
                        raise ValueError("Invalid object type: " + target + ".")
                    obj = _GXObjectFactory.createObject(type_)
                    obj.version = 0
                    reader.objects.append(obj)
                    reader.getNext()
                elif "SN".lower() == target.lower():
                    obj.shortName = reader.readElementContentAsInt("SN")
                elif "LN".lower() == target.lower():
                    obj.logicalName = reader.readElementContentAsString("LN")
                elif "Description".lower() == target.lower():
                    obj.description = reader.readElementContentAsString("Description")
                elif "Version".lower() == target.lower():
                    obj.version = reader.readElementContentAsInt("Version")
                else:
                    # pylint:disable=broad-except
                    try:
                        obj.load(reader)
                    except Exception as ex:
                        print("Failed to load object " + str(obj) + " " +str(ex))
                    obj = None
            else:
                reader.read()
        for obj in reader.objects:
            obj.postLoad(reader)
        return reader.objects

    def save(self, name, settings=None):
        writer = GXXmlWriter()
        objects = ET.Element("Objects")
        for it in self:
            node = ET.SubElement(objects, "GXDLMS" + GXDLMSConverter.objectTypeToString(it.objectType))
            if it.shortName != 0:
                ET.SubElement(node, "SN").text = str(it.shortName)
            ET.SubElement(node, "LN").text = it.logicalName
            if it.version != 0:
                ET.SubElement(node, "Version").text = str(it.version)
            if it.description:
                ET.SubElement(node, "Description").text = it.description
            if not settings or settings.values:
                writer.objects = []
                writer.objects.append(node)
                it.save(writer)
        str_ = minidom.parseString(ET.tostring(objects, encoding='utf-8', method='xml')).toprettyxml(indent="  ")
        with open(name, "w") as f:
            f.write(str_)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class,too-few-public-methods
class GXDLMSModemInitialisation:
    #
    #Constructor.
    #
    def __init__(self):
        self.request = None
        self.response = None
        self.delay = 0

    def __str__(self):
        return self.request + " " + self.response + " " + str(self.delay)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSEmergencyProfile import GXDLMSEmergencyProfile
from .GXDLMSActionItem import GXDLMSActionItem

# pylint: disable=too-many-instance-attributes
class GXDLMSLimiter(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSLimiter

    """
    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        #pylint: disable=super-with-arguments
        super(GXDLMSLimiter, self).__init__(ObjectType.LIMITER, ln, sn)
        self.monitoredValue = None
        self.monitoredAttributeIndex = 0
        self.thresholdActive = None
        self.thresholdNormal = None
        self.thresholdEmergency = None
        self.minOverThresholdDuration = None
        self.minUnderThresholdDuration = None
        self.emergencyProfileGroupIDs = list()
        self.emergencyProfileActive = None
        self.emergencyProfile = GXDLMSEmergencyProfile()
        self.actionOverThreshold = GXDLMSActionItem()
        self.actionUnderThreshold = GXDLMSActionItem()

    def getValues(self):
        return [self.logicalName,
                self.monitoredValue,
                self.thresholdActive,
                self.thresholdNormal,
                self.thresholdEmergency,
                self.minOverThresholdDuration,
                self.minUnderThresholdDuration,
                self.emergencyProfile,
                self.emergencyProfileGroupIDs,
                self.emergencyProfileActive,
                [self.actionOverThreshold, self.actionUnderThreshold]]

    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  MonitoredValue
        if all_ or self.canRead(2):
            attributes.append(2)
        #  ThresholdActive
        if all_ or self.canRead(3):
            attributes.append(3)
        #  ThresholdNormal
        if all_ or self.canRead(4):
            attributes.append(4)
        #  ThresholdEmergency
        if all_ or self.canRead(5):
            attributes.append(5)
        #  MinOverThresholdDuration
        if all_ or self.canRead(6):
            attributes.append(6)
        #  MinUnderThresholdDuration
        if all_ or self.canRead(7):
            attributes.append(7)
        #  EmergencyProfile
        if all_ or self.canRead(8):
            attributes.append(8)
        #  EmergencyProfileGroup
        if all_ or self.canRead(9):
            attributes.append(9)
        #  EmergencyProfileActive
        if all_ or self.canRead(10):
            attributes.append(10)
        #  Actions
        if all_ or self.canRead(11):
            attributes.append(11)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 11

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        #pylint: disable=super-with-arguments
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.STRUCTURE
        elif index == 3:
            ret = super(GXDLMSLimiter, self).getDataType(index)
        elif index == 4:
            ret = super(GXDLMSLimiter, self).getDataType(index)
        elif index == 5:
            ret = super(GXDLMSLimiter, self).getDataType(index)
        elif index == 6:
            ret = DataType.UINT32
        elif index == 7:
            ret = DataType.UINT32
        elif index == 8:
            ret = DataType.STRUCTURE
        elif index == 9:
            ret = DataType.ARRAY
        elif index == 10:
            ret = DataType.BOOLEAN
        elif index == 11:
            ret = DataType.STRUCTURE
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(3)
            if self.monitoredValue is None:
                _GXCommon.setData(settings, data, DataType.UINT16, int(0))
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(None))
                _GXCommon.setData(settings, data, DataType.INT8, 0)
            else:
                _GXCommon.setData(settings, data, DataType.UINT16, int(self.monitoredValue.objectType))
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(self.monitoredValue.logicalName))
                _GXCommon.setData(settings, data, DataType.INT8, self.monitoredAttributeIndex)
            ret = data
        elif e.index == 3:
            ret = self.thresholdActive
        elif e.index == 4:
            ret = self.thresholdNormal
        elif e.index == 5:
            ret = self.thresholdEmergency
        elif e.index == 6:
            ret = self.minOverThresholdDuration
        elif e.index == 7:
            ret = self.minUnderThresholdDuration
        elif e.index == 8:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(3)
            _GXCommon.setData(settings, data, DataType.UINT16, self.emergencyProfile.id)
            _GXCommon.setData(settings, data, DataType.OCTET_STRING, self.emergencyProfile.activationTime)
            _GXCommon.setData(settings, data, DataType.UINT32, self.emergencyProfile.duration)
            ret = data
        elif e.index == 9:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            data.setUInt8(len(self.emergencyProfileGroupIDs))
            for it in self.emergencyProfileGroupIDs:
                _GXCommon.setData(settings, data, DataType.UINT16, it)
            ret = data
        elif e.index == 10:
            ret = self.emergencyProfileActive
        elif e.index == 11:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(self.actionOverThreshold.logicalName))
            _GXCommon.setData(settings, data, DataType.UINT16, self.actionOverThreshold.scriptSelector)
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(self.actionUnderThreshold.logicalName))
            _GXCommon.setData(settings, data, DataType.UINT16, self.actionUnderThreshold.scriptSelector)
            ret = data
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            ot = e.value[0]
            ln = _GXCommon.toLogicalName(e.value[1])
            self.monitoredAttributeIndex = e.value[2]
            self.monitoredValue = settings.objects.findByLN(ot, ln)
        elif e.index == 3:
            self.thresholdActive = e.value
        elif e.index == 4:
            self.thresholdNormal = e.value
        elif e.index == 5:
            self.thresholdEmergency = e.value
        elif e.index == 6:
            self.minOverThresholdDuration = e.value
        elif e.index == 7:
            self.minUnderThresholdDuration = e.value
        elif e.index == 8:
            self.emergencyProfile.id = e.value[0]
            self.emergencyProfile.activationTime = _GXCommon.changeType(settings, e.value[1], DataType.DATETIME)
            self.emergencyProfile.duration = e.value[2]
        elif e.index == 9:
            self.emergencyProfileGroupIDs = []
            if e.value:
                for it in e.value:
                    self.emergencyProfileGroupIDs.append(it)
        elif e.index == 10:
            self.emergencyProfileActive = e.value
        elif e.index == 11:
            self.actionOverThreshold.logicalName = _GXCommon.toLogicalName(e.value[0][0])
            self.actionOverThreshold.scriptSelector = e.value[0][1]
            self.actionUnderThreshold.logicalName = _GXCommon.toLogicalName(e.value[1][0])
            self.actionUnderThreshold.scriptSelector = e.value[1][1]
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        #pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory
        if reader.isStartElement("MonitoredValue", True):
            ot = reader.readElementContentAsInt("ObjectType")
            ln = reader.readElementContentAsString("LN")
            if ot != ObjectType.NONE and ln:
                self.monitoredValue = reader.objects.findByLN(ot, ln)
                #  If item is not serialized yet.
                if self.monitoredValue is None:
                    self.monitoredValue = _GXObjectFactory.createObject(ot)
                    self.monitoredValue.logicalName = ln
            reader.readEndElement("MonitoredValue")
        self.thresholdActive = reader.readElementContentAsObject("ThresholdActive", None, self, 3)
        self.thresholdNormal = reader.readElementContentAsObject("ThresholdNormal", None, self, 4)
        self.thresholdEmergency = reader.readElementContentAsObject("ThresholdEmergency", None, self, 5)
        self.minOverThresholdDuration = reader.readElementContentAsInt("MinOverThresholdDuration")
        self.minUnderThresholdDuration = reader.readElementContentAsInt("MinUnderThresholdDuration")
        if reader.isStartElement("EmergencyProfile", True):
            self.emergencyProfile.id = reader.readElementContentAsInt("ID")
            self.emergencyProfile.activationTime = reader.readElementContentAsDateTime("Time")
            self.emergencyProfile.duration = reader.readElementContentAsInt("Duration")
            reader.readEndElement("EmergencyProfile")
        self.emergencyProfileGroupIDs = []
        if reader.isStartElement("EmergencyProfileGroupIDs", True):
            while reader.isStartElement("Value", False):
                self.emergencyProfileGroupIDs.append(reader.readElementContentAsInt("Value"))
            reader.readEndElement("EmergencyProfileGroupIDs")
        self.emergencyProfileActive = reader.readElementContentAsInt("Active") != 0
        if reader.isStartElement("ActionOverThreshold", True):
            self.actionOverThreshold.logicalName = reader.readElementContentAsString("LN")
            self.actionOverThreshold.scriptSelector = reader.readElementContentAsInt("ScriptSelector")
            reader.readEndElement("ActionOverThreshold")
        if reader.isStartElement("ActionUnderThreshold", True):
            self.actionUnderThreshold.logicalName = reader.readElementContentAsString("LN")
            self.actionUnderThreshold.scriptSelector = reader.readElementContentAsInt("ScriptSelector")
            reader.readEndElement("ActionUnderThreshold")

    def save(self, writer):
        if self.monitoredValue:
            writer.writeStartElement("MonitoredValue")
            writer.writeElementString("ObjectType", int(self.monitoredValue.objectType))
            writer.writeElementString("LN", self.monitoredValue.logicalName)
            writer.writeEndElement()
        writer.writeElementObject("ThresholdActive", self.thresholdActive, self.getDataType(3), self.getUIDataType(3))
        writer.writeElementObject("ThresholdNormal", self.thresholdNormal, self.getDataType(4), self.getUIDataType(4))
        writer.writeElementObject("ThresholdEmergency", self.thresholdEmergency, self.getDataType(5), self.getUIDataType(5))
        writer.writeElementString("MinOverThresholdDuration", self.minOverThresholdDuration)
        writer.writeElementString("MinUnderThresholdDuration", self.minUnderThresholdDuration)
        if self.emergencyProfile:
            writer.writeStartElement("EmergencyProfile")
            writer.writeElementString("ID", self.emergencyProfile.id)
            writer.writeElementString("Time", self.emergencyProfile.activationTime)
            writer.writeElementString("Duration", self.emergencyProfile.duration)
            writer.writeEndElement()
        writer.writeStartElement("EmergencyProfileGroupIDs")
        if self.emergencyProfileGroupIDs:
            for it in self.emergencyProfileGroupIDs:
                writer.writeElementString("Value", it)
        writer.writeEndElement()
        writer.writeElementString("Active", self.emergencyProfileActive)
        if self.actionOverThreshold:
            writer.writeStartElement("ActionOverThreshold")
            writer.writeElementString("LN", self.actionOverThreshold.logicalName)
            writer.writeElementString("ScriptSelector", self.actionOverThreshold.scriptSelector)
            writer.writeEndElement()
        if self.actionUnderThreshold:
            writer.writeStartElement("ActionUnderThreshold")
            writer.writeElementString("LN", self.actionUnderThreshold.logicalName)
            writer.writeElementString("ScriptSelector", self.actionUnderThreshold.scriptSelector)
            writer.writeEndElement()

    def postLoad(self, reader):
        #  Upload Monitored Value after load.
        if self.monitoredValue:
            target = reader.objects.findByLN(self.monitoredValue.objectType, self.monitoredValue.logicalName)
            if target != self.monitoredValue:
                self.monitoredValue = target



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType

# pylint: disable=too-many-instance-attributes
class GXDLMSPrimeNbOfdmPlcMacSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPrimeNbOfdmPlcMacSetup
    """

    def __init__(self, ln="0.0.28.2.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.PRIME_NB_OFDM_PLC_MAC_SETUP, ln, sn)
        # PIB attribute 0x0010.
        self.macMinSwitchSearchTime = 0
        # PIB attribute 0x0011.
        self.macMaxPromotionPdu = 0
        # PIB attribute 0x0012.
        self.macPromotionPduTxPeriod = 0
        # PIB attribute 0x0013.
        self.macBeaconsPerFrame = 0
        # PIB attribute 0x0014.
        self.macScpMaxTxAttempts = 0
        # PIB attribute 0x0015.
        self.macCtlReTxTimer = 0
        # PIB attribute 0x0018.
        self.macMaxCtlReTx = 0

    def getValues(self):
        return [self.logicalName,
                self.macMinSwitchSearchTime,
                self.macMaxPromotionPdu,
                self.macPromotionPduTxPeriod,
                self.macBeaconsPerFrame,
                self.macScpMaxTxAttempts,
                self.macCtlReTxTimer,
                self.macMaxCtlReTx]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  MacMinSwitchSearchTime
        if all_ or self.canRead(2):
            attributes.append(2)
        #  MacMaxPromotionPdu
        if all_ or self.canRead(3):
            attributes.append(3)
        #  MacPromotionPduTxPeriod
        if all_ or self.canRead(4):
            attributes.append(4)
        #  MacBeaconsPerFrame
        if all_ or self.canRead(5):
            attributes.append(5)
        #  MacScpMaxTxAttempts
        if all_ or self.canRead(6):
            attributes.append(6)
        #  MacCtlReTxTimer
        if all_ or self.canRead(7):
            attributes.append(7)
        #  MacMaxCtlReTx
        if all_ or self.canRead(8):
            attributes.append(8)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 8

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index in (2, 3, 4, 5, 6, 7, 8):
            return DataType.UINT8
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.macMinSwitchSearchTime
        elif e.index == 3:
            ret = self.macMaxPromotionPdu
        elif e.index == 4:
            ret = self.macPromotionPduTxPeriod
        elif e.index == 5:
            ret = self.macBeaconsPerFrame
        elif e.index == 6:
            ret = self.macScpMaxTxAttempts
        elif e.index == 7:
            ret = self.macCtlReTxTimer
        elif e.index == 8:
            ret = self.macMaxCtlReTx
        else:
            ret = None
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.macMinSwitchSearchTime = e.value
        elif e.index == 3:
            self.macMaxPromotionPdu = e.value
        elif e.index == 4:
            self.macPromotionPduTxPeriod = e.value
        elif e.index == 5:
            self.macBeaconsPerFrame = e.value
        elif e.index == 6:
            self.macScpMaxTxAttempts = e.value
        elif e.index == 7:
            self.macCtlReTxTimer = e.value
        elif e.index == 8:
            self.macMaxCtlReTx = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.macMinSwitchSearchTime = reader.readElementContentAsInt("MacMinSwitchSearchTime")
        self.macMaxPromotionPdu = reader.readElementContentAsInt("MacMaxPromotionPdu")
        self.macPromotionPduTxPeriod = reader.readElementContentAsInt("MacPromotionPduTxPeriod")
        self.macBeaconsPerFrame = reader.readElementContentAsInt("MacBeaconsPerFrame")
        self.macScpMaxTxAttempts = reader.readElementContentAsInt("MacScpMaxTxAttempts")
        self.macCtlReTxTimer = reader.readElementContentAsInt("MacCtlReTxTimer")
        self.macMaxCtlReTx = reader.readElementContentAsInt("MacMaxCtlReTx")

    def save(self, writer):
        writer.writeElementString("MacMinSwitchSearchTime", self.macMinSwitchSearchTime)
        writer.writeElementString("MacMaxPromotionPdu", self.macMaxPromotionPdu)
        writer.writeElementString("MacPromotionPduTxPeriod", self.macPromotionPduTxPeriod)
        writer.writeElementString("MacBeaconsPerFrame", self.macBeaconsPerFrame)
        writer.writeElementString("MacScpMaxTxAttempts", self.macScpMaxTxAttempts)
        writer.writeElementString("MacCtlReTxTimer", self.macCtlReTxTimer)
        writer.writeElementString("MacMaxCtlReTx", self.macMaxCtlReTx)
